from decimal import Decimal
from unittest import mock

import pytest
from yarl import URL

from neuro_config_client.converters import PrimitiveToClusterConverter
from neuro_config_client.models import (
    ACMEEnvironment,
    ARecord,
    BlobStorageConfig,
    BucketsConfig,
    Cluster,
    DisksConfig,
    DNSConfig,
    IdleJobConfig,
    IngressConfig,
    MetricsConfig,
    MonitoringConfig,
    OrchestratorConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)


class TestPrimitiveToCLusterConverter:
    @pytest.fixture
    def converter(self) -> PrimitiveToClusterConverter:
        return PrimitiveToClusterConverter()

    def test_convert_empty_cluster(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_cluster({"name": "default"})

        assert result == Cluster(name="default")

    def test_convert_cluster(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_cluster(
            {
                "name": "default",
                "orchestrator": {
                    "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                    "job_fallback_hostname": "default.jobs-dev.neu.ro",
                    "job_schedule_timeout_s": 1,
                    "job_schedule_scale_up_timeout_s": 2,
                    "is_http_ingress_secure": False,
                    "resource_pool_types": [{"name": "node-pool"}],
                    "allow_privileged_mode": False,
                },
                "storage": {"url": "https://storage-dev.neu.ro"},
                "registry": {
                    "url": "https://registry-dev.neu.ro",
                    "email": "dev@neu.ro",
                },
                "monitoring": {"url": "https://monitoring-dev.neu.ro"},
                "secrets": {"url": "https://secrets-dev.neu.ro"},
                "metrics": {"url": "https://secrets-dev.neu.ro"},
                "disks": {
                    "url": "https://secrets-dev.neu.ro",
                    "storage_limit_per_user_gb": 1024,
                },
                "ingress": {"acme_environment": "production"},
                "dns": {
                    "name": "neu.ro",
                    "a_records": [
                        {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
                    ],
                },
            }
        )

        assert result.name == "default"
        assert result.orchestrator
        assert result.storage
        assert result.registry
        assert result.monitoring
        assert result.secrets
        assert result.metrics
        assert result.disks
        assert result.ingress
        assert result.dns

    def test_convert_orchestrator(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_internal_hostname_template": "{job_id}.platform-jobs",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
                "resource_pool_types": [{"name": "node-pool"}],
                "resource_presets": [
                    {
                        "name": "cpu-micro",
                        "credits_per_hour": "10",
                        "cpu": 0.1,
                        "memory_mb": 100,
                    }
                ],
                "allow_privileged_mode": False,
                "pre_pull_images": ["neuromation/base"],
                "idle_jobs": [
                    {
                        "count": 1,
                        "image": "miner",
                        "resources": {"cpu_m": 1000, "memory_mb": 1024},
                    }
                ],
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_internal_hostname_template="{job_id}.platform-jobs",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            resource_pool_types=[mock.ANY],
            resource_presets=[mock.ANY],
            allow_privileged_mode=False,
            pre_pull_images=["neuromation/base"],
            idle_jobs=[
                IdleJobConfig(
                    count=1,
                    image="miner",
                    resources=Resources(cpu_m=1000, memory_mb=1024),
                )
            ],
        )

    def test_convert_orchestrator_default(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_orchestrator(
            {
                "job_hostname_template": "{job_id}.jobs-dev.neu.ro",
                "job_fallback_hostname": "default.jobs-dev.neu.ro",
                "job_schedule_timeout_s": 1,
                "job_schedule_scale_up_timeout_s": 2,
                "is_http_ingress_secure": False,
                "allow_privileged_mode": False,
            }
        )

        assert result == OrchestratorConfig(
            job_hostname_template="{job_id}.jobs-dev.neu.ro",
            job_internal_hostname_template="",
            job_fallback_hostname="default.jobs-dev.neu.ro",
            job_schedule_timeout_s=1,
            job_schedule_scale_up_timeout_s=2,
            is_http_ingress_secure=False,
            allow_privileged_mode=False,
        )

    def test_convert_resource_pool_type(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_resource_pool_type(
            {
                "name": "n1-highmem-4",
                "min_size": 1,
                "max_size": 2,
                "idle_size": 1,
                "cpu": 4.0,
                "available_cpu": 3.0,
                "memory_mb": 12 * 1024,
                "available_memory_mb": 10 * 1024,
                "disk_size_gb": 700,
                "gpu": 1,
                "gpu_model": "nvidia-tesla-k80",
                "tpu": {
                    "ipv4_cidr_block": "10.0.0.0/8",
                    "types": ["tpu"],
                    "software_versions": ["v1"],
                },
                "is_preemptible": True,
                "price": "1.0",
                "currency": "USD",
            }
        )

        assert result == ResourcePoolType(
            name="n1-highmem-4",
            min_size=1,
            max_size=2,
            idle_size=1,
            cpu=4.0,
            available_cpu=3.0,
            memory_mb=12 * 1024,
            available_memory_mb=10 * 1024,
            gpu=1,
            gpu_model="nvidia-tesla-k80",
            tpu=mock.ANY,
            is_preemptible=True,
            price=Decimal("1.0"),
            currency="USD",
        )

    def test_convert_empty_resource_pool_type(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_resource_pool_type({"name": "node-pool"})

        assert result == ResourcePoolType(name="node-pool")

    def test_convert_tpu_resource(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_tpu_resource(
            {
                "ipv4_cidr_block": "10.0.0.0/8",
                "types": ["tpu"],
                "software_versions": ["v1"],
            }
        )

        assert result == TPUResource(
            ipv4_cidr_block="10.0.0.0/8", types=["tpu"], software_versions=["v1"]
        )

    def test_convert_resource_preset(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_resource_preset(
            {
                "name": "cpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory_mb": 1024,
            }
        )

        assert result == ResourcePreset(
            name="cpu-small", credits_per_hour=Decimal("10"), cpu=4.0, memory_mb=1024
        )

    def test_convert_resource_preset_with_memory_gpu_tpu_preemptible_affinity(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_resource_preset(
            {
                "name": "gpu-small",
                "credits_per_hour": "10",
                "cpu": 4.0,
                "memory_mb": 12288,
                "gpu": 1,
                "gpu_model": "nvidia-tesla-k80",
                "tpu": {"type": "tpu", "software_version": "v1"},
                "scheduler_enabled": True,
                "preemptible_node": True,
                "resource_affinity": ["gpu-k80"],
            }
        )

        assert result == ResourcePreset(
            name="gpu-small",
            credits_per_hour=Decimal("10"),
            cpu=4.0,
            memory_mb=12288,
            gpu=1,
            gpu_model="nvidia-tesla-k80",
            tpu=TPUPreset(type="tpu", software_version="v1"),
            scheduler_enabled=True,
            preemptible_node=True,
            resource_affinity=["gpu-k80"],
        )

    def test_convert_storage(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_storage({"url": "https://storage-dev.neu.ro"})

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"), volumes=[]
        )

    def test_convert_storage_with_volumes(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_storage(
            {
                "url": "https://storage-dev.neu.ro",
                "volumes": [
                    {},
                    {"path": "/volume", "size_mb": 1024},
                ],
            }
        )

        assert result == StorageConfig(
            url=URL("https://storage-dev.neu.ro"),
            volumes=[
                VolumeConfig(),
                VolumeConfig(path="/volume", size_mb=1024),
            ],
        )

    def test_convert_blob_storage(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_blob_storage(
            {"url": "https://blob-storage-dev.neu.ro"}
        )

        assert result == BlobStorageConfig(url=URL("https://blob-storage-dev.neu.ro"))

    def test_convert_registry(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_registry({"url": "https://registry-dev.neu.ro"})

        assert result == RegistryConfig(url=URL("https://registry-dev.neu.ro"))

    def test_convert_monitoring(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_monitoring({"url": "https://monitoring-dev.neu.ro"})

        assert result == MonitoringConfig(url=URL("https://monitoring-dev.neu.ro"))

    def test_convert_secrets(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_secrets({"url": "https://secrets-dev.neu.ro"})

        assert result == SecretsConfig(url=URL("https://secrets-dev.neu.ro"))

    def test_convert_metrics(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_metrics({"url": "https://metrics-dev.neu.ro"})

        assert result == MetricsConfig(url=URL("https://metrics-dev.neu.ro"))

    def test_convert_dns(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_dns(
            {
                "name": "neu.ro",
                "a_records": [{"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}],
            }
        )

        assert result == DNSConfig(name="neu.ro", a_records=[mock.ANY])

    def test_convert_a_record_with_ips(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_a_record(
            {"name": "*.jobs-dev.neu.ro.", "ips": ["192.168.0.2"]}
        )

        assert result == ARecord(name="*.jobs-dev.neu.ro.", ips=["192.168.0.2"])

    def test_convert_a_record_dns_name(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_a_record(
            {
                "name": "*.jobs-dev.neu.ro.",
                "dns_name": "load-balancer",
                "zone_id": "/hostedzone/1",
                "evaluate_target_health": True,
            }
        )

        assert result == ARecord(
            name="*.jobs-dev.neu.ro.",
            dns_name="load-balancer",
            zone_id="/hostedzone/1",
            evaluate_target_health=True,
        )

    def test_convert_disks(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_disks(
            {"url": "https://metrics-dev.neu.ro", "storage_limit_per_user_gb": 1024}
        )

        assert result == DisksConfig(
            url=URL("https://metrics-dev.neu.ro"), storage_limit_per_user_gb=1024
        )

    def test_convert_buckets(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_buckets(
            {"url": "https://buckets-dev.neu.ro", "disable_creation": True}
        )

        assert result == BucketsConfig(
            url=URL("https://buckets-dev.neu.ro"), disable_creation=True
        )

    def test_convert_ingress(self, converter: PrimitiveToClusterConverter) -> None:
        result = converter.convert_ingress(
            {"acme_environment": "production", "cors_origins": ["https://app.neu.ro"]}
        )

        assert result == IngressConfig(
            acme_environment=ACMEEnvironment.PRODUCTION,
            cors_origins=["https://app.neu.ro"],
        )

    def test_convert_ingress_defaults(
        self, converter: PrimitiveToClusterConverter
    ) -> None:
        result = converter.convert_ingress({"acme_environment": "production"})

        assert result == IngressConfig(acme_environment=ACMEEnvironment.PRODUCTION)



import enum
from collections.abc import Sequence
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Optional

from yarl import URL


@dataclass(frozen=True)
class VolumeConfig:
    size_mb: Optional[int] = None
    path: Optional[str] = None


@dataclass(frozen=True)
class StorageConfig:
    url: URL
    volumes: Sequence[VolumeConfig] = ()


@dataclass(frozen=True)
class BlobStorageConfig:
    url: URL


@dataclass(frozen=True)
class RegistryConfig:
    url: URL


@dataclass(frozen=True)
class MonitoringConfig:
    url: URL


@dataclass(frozen=True)
class MetricsConfig:
    url: URL


@dataclass(frozen=True)
class SecretsConfig:
    url: URL


@dataclass(frozen=True)
class DisksConfig:
    url: URL
    storage_limit_per_user_gb: int


@dataclass(frozen=True)
class BucketsConfig:
    url: URL
    disable_creation: bool = False


class ACMEEnvironment(str, enum.Enum):
    STAGING = "staging"
    PRODUCTION = "production"


@dataclass(frozen=True)
class IngressConfig:
    acme_environment: ACMEEnvironment
    cors_origins: Sequence[str] = ()


@dataclass(frozen=True)
class TPUResource:
    ipv4_cidr_block: str
    types: Sequence[str] = field(default_factory=list)
    software_versions: Sequence[str] = field(default_factory=list)


@dataclass(frozen=True)
class TPUPreset:
    type: str
    software_version: str


@dataclass(frozen=True)
class ResourcePreset:
    name: str
    credits_per_hour: Decimal
    cpu: float
    memory_mb: int
    gpu: Optional[int] = None
    gpu_model: Optional[str] = None
    tpu: Optional[TPUPreset] = None
    scheduler_enabled: bool = False
    preemptible_node: bool = False
    resource_affinity: Sequence[str] = ()


@dataclass(frozen=True)
class ResourcePoolType:
    name: str
    min_size: int = 0
    max_size: int = 1
    idle_size: int = 0
    cpu: float = 1.0
    available_cpu: float = 1.0
    memory_mb: int = 1024
    available_memory_mb: int = 1024
    disk_size_gb: int = 150
    gpu: Optional[int] = None
    gpu_model: Optional[str] = None
    price: Decimal = Decimal()
    currency: Optional[str] = None
    tpu: Optional[TPUResource] = None
    is_preemptible: bool = False


@dataclass(frozen=True)
class Resources:
    cpu_m: int
    memory_mb: int
    gpu: int = 0


@dataclass(frozen=True)
class IdleJobConfig:
    count: int
    image: str
    resources: Resources
    image_secret: str = ""
    env: dict[str, str] = field(default_factory=dict)
    node_selector: dict[str, str] = field(default_factory=dict)


@dataclass
class OrchestratorConfig:
    job_hostname_template: str
    job_internal_hostname_template: str
    job_fallback_hostname: str
    job_schedule_timeout_s: float
    job_schedule_scale_up_timeout_s: float
    is_http_ingress_secure: bool = True
    resource_pool_types: Sequence[ResourcePoolType] = field(default_factory=list)
    resource_presets: Sequence[ResourcePreset] = field(default_factory=list)
    allow_privileged_mode: bool = False
    pre_pull_images: Sequence[str] = ()
    idle_jobs: Sequence[IdleJobConfig] = field(default_factory=list)


@dataclass
class ARecord:
    name: str
    ips: Sequence[str] = field(default_factory=list)
    dns_name: str = ""
    zone_id: str = ""
    evaluate_target_health: bool = False


@dataclass
class DNSConfig:
    name: str
    a_records: Sequence[ARecord] = field(default_factory=list)


@dataclass(frozen=True)
class Cluster:
    name: str
    orchestrator: Optional[OrchestratorConfig] = None
    storage: Optional[StorageConfig] = None
    blob_storage: Optional[BlobStorageConfig] = None
    registry: Optional[RegistryConfig] = None
    monitoring: Optional[MonitoringConfig] = None
    secrets: Optional[SecretsConfig] = None
    metrics: Optional[MetricsConfig] = None
    dns: Optional[DNSConfig] = None
    disks: Optional[DisksConfig] = None
    buckets: Optional[BucketsConfig] = None
    ingress: Optional[IngressConfig] = None



from __future__ import annotations

import logging
from collections.abc import Sequence
from types import TracebackType
from typing import Any

import aiohttp
from aiohttp import ClientResponseError
from yarl import URL

from .converters import PrimitiveToClusterConverter
from .models import Cluster

logger = logging.getLogger(__name__)


class ConfigClient:
    def __init__(
        self,
        url: URL,
        token: str,
        timeout: aiohttp.ClientTimeout = aiohttp.client.DEFAULT_TIMEOUT,
        trace_configs: Sequence[aiohttp.TraceConfig] = (),
    ):
        self._clusters_url = url / "api/v1/clusters"
        self._token = token
        self._timeout = timeout
        self._trace_configs = trace_configs
        self._client: aiohttp.ClientSession | None = None
        self._primitive_to_cluster_converter = PrimitiveToClusterConverter()

    async def __aenter__(self) -> "ConfigClient":
        self._client = await self._create_http_client()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.aclose()

    async def aclose(self) -> None:
        assert self._client
        await self._client.close()

    async def _create_http_client(self) -> aiohttp.ClientSession:
        client = aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {self._token}"},
            timeout=self._timeout,
            trace_configs=list(self._trace_configs),
        )
        return await client.__aenter__()

    async def get_clusters(self) -> Sequence[Cluster]:
        assert self._client
        async with self._client.get(
            self._clusters_url, params={"include": "all"}
        ) as response:
            response.raise_for_status()
            payload = await response.json()
            return [
                self._primitive_to_cluster_converter.convert_cluster(p) for p in payload
            ]

    async def get_cluster(self, name: str) -> Cluster:
        assert self._client
        async with self._client.get(
            self._clusters_url / name, params={"include": "all"}
        ) as response:
            response.raise_for_status()
            payload = await response.json()
            return self._primitive_to_cluster_converter.convert_cluster(payload)

    async def create_blank_cluster(
        self, name: str, token: str, ignore_existing: bool = False
    ) -> None:
        assert self._client
        payload = {"name": name, "token": token}
        try:
            async with self._client.post(self._clusters_url, json=payload) as resp:
                resp.raise_for_status()
        except ClientResponseError as e:
            is_existing = e.status == 400 and "already exists" in e.message
            if not ignore_existing or is_existing:
                raise

    async def delete_cluster(self, name: str) -> None:
        assert self._client
        async with self._client.delete(self._clusters_url / name) as resp:
            resp.raise_for_status()

    async def add_storage(
        self,
        cluster_name: str,
        storage_name: str,
        size_mb: int | None = None,
        *,
        start_deployment: bool = True,
        ignore_existing: bool = False,
    ) -> None:
        assert self._client
        try:
            url = self._clusters_url / cluster_name / "cloud_provider/storages"
            payload: dict[str, Any] = {"name": storage_name}
            if size_mb is not None:
                payload["size_mb"] = size_mb
            async with self._client.post(
                url.with_query(start_deployment=str(start_deployment).lower()),
                json=payload,
            ) as response:
                response.raise_for_status()
        except ClientResponseError as e:
            if not ignore_existing or e.status != 409:
                raise

    async def remove_storage(
        self,
        cluster_name: str,
        storage_name: str,
        *,
        start_deployment: bool = True,
        ignore_not_found: bool = False,
    ) -> None:
        assert self._client
        try:
            url = (
                self._clusters_url
                / cluster_name
                / "cloud_provider/storages"
                / storage_name
            )
            async with self._client.delete(
                url.with_query(start_deployment=str(start_deployment).lower())
            ) as response:
                response.raise_for_status()
        except ClientResponseError as e:
            if not ignore_not_found or e.status != 404:
                raise



from decimal import Decimal
from typing import Any

from yarl import URL

from .models import (
    ACMEEnvironment,
    ARecord,
    BlobStorageConfig,
    BucketsConfig,
    Cluster,
    DisksConfig,
    DNSConfig,
    IdleJobConfig,
    IngressConfig,
    MetricsConfig,
    MonitoringConfig,
    OrchestratorConfig,
    RegistryConfig,
    ResourcePoolType,
    ResourcePreset,
    Resources,
    SecretsConfig,
    StorageConfig,
    TPUPreset,
    TPUResource,
    VolumeConfig,
)


class PrimitiveToClusterConverter:
    def convert_cluster(self, payload: dict[str, Any]) -> Cluster:
        orchestrator = None
        if payload.get("orchestrator"):
            orchestrator = self.convert_orchestrator(payload["orchestrator"])
        storage = None
        if payload.get("storage"):
            storage = self.convert_storage(payload["storage"])
        blob_storage = None
        if payload.get("blob_storage"):
            blob_storage = self.convert_blob_storage(payload["blob_storage"])
        registry = None
        if payload.get("registry"):
            registry = self.convert_registry(payload["registry"])
        monitoring = None
        if payload.get("monitoring"):
            monitoring = self.convert_monitoring(payload["monitoring"])
        secrets = None
        if payload.get("secrets"):
            secrets = self.convert_secrets(payload["secrets"])
        metrics = None
        if payload.get("metrics"):
            metrics = self.convert_metrics(payload["metrics"])
        disks = None
        if payload.get("disks"):
            disks = self.convert_disks(payload["disks"])
        buckets = None
        if payload.get("buckets"):
            buckets = self.convert_buckets(payload["buckets"])
        ingress = None
        if payload.get("ingress"):
            ingress = self.convert_ingress(payload["ingress"])
        dns = None
        if payload.get("dns"):
            dns = self.convert_dns(payload["dns"])
        return Cluster(
            name=payload["name"],
            orchestrator=orchestrator,
            storage=storage,
            blob_storage=blob_storage,
            registry=registry,
            monitoring=monitoring,
            secrets=secrets,
            metrics=metrics,
            disks=disks,
            buckets=buckets,
            ingress=ingress,
            dns=dns,
        )

    def convert_orchestrator(self, payload: dict[str, Any]) -> OrchestratorConfig:
        return OrchestratorConfig(
            job_hostname_template=payload["job_hostname_template"],
            job_internal_hostname_template=payload.get(
                "job_internal_hostname_template", ""
            ),
            job_fallback_hostname=payload["job_fallback_hostname"],
            job_schedule_timeout_s=payload["job_schedule_timeout_s"],
            job_schedule_scale_up_timeout_s=payload["job_schedule_scale_up_timeout_s"],
            is_http_ingress_secure=payload["is_http_ingress_secure"],
            resource_pool_types=[
                self.convert_resource_pool_type(r)
                for r in payload.get("resource_pool_types", [])
            ],
            resource_presets=[
                self.convert_resource_preset(preset)
                for preset in payload.get("resource_presets", [])
            ],
            allow_privileged_mode=payload["allow_privileged_mode"],
            pre_pull_images=payload.get("pre_pull_images", ()),
            idle_jobs=[
                self.convert_idle_job(job) for job in payload.get("idle_jobs", ())
            ],
        )

    def convert_resource_pool_type(self, payload: dict[str, Any]) -> ResourcePoolType:
        tpu = None
        if payload.get("tpu"):
            tpu = self.convert_tpu_resource(payload["tpu"])
        return ResourcePoolType(
            name=payload["name"],
            min_size=payload.get("min_size", ResourcePoolType.min_size),
            max_size=payload.get("max_size", ResourcePoolType.max_size),
            idle_size=payload.get("idle_size", ResourcePoolType.idle_size),
            cpu=payload.get("cpu", ResourcePoolType.cpu),
            available_cpu=payload.get("available_cpu", ResourcePoolType.available_cpu),
            memory_mb=payload.get("memory_mb", ResourcePoolType.memory_mb),
            available_memory_mb=payload.get(
                "available_memory_mb", ResourcePoolType.available_memory_mb
            ),
            gpu=payload.get("gpu"),
            gpu_model=payload.get("gpu_model"),
            price=Decimal(payload.get("price", ResourcePoolType.price)),
            currency=payload.get("currency"),
            tpu=tpu,
            is_preemptible=payload.get(
                "is_preemptible", ResourcePoolType.is_preemptible
            ),
        )

    def convert_tpu_resource(self, payload: dict[str, Any]) -> TPUResource:
        return TPUResource(
            ipv4_cidr_block=payload["ipv4_cidr_block"],
            types=list(payload["types"]),
            software_versions=list(payload["software_versions"]),
        )

    def convert_resource_preset(self, payload: dict[str, Any]) -> ResourcePreset:
        tpu = None
        if payload.get("tpu"):
            tpu = self.convert_tpu_preset(payload["tpu"])
        return ResourcePreset(
            name=payload["name"],
            credits_per_hour=Decimal(payload["credits_per_hour"]),
            cpu=payload["cpu"],
            memory_mb=payload["memory_mb"],
            gpu=payload.get("gpu"),
            gpu_model=payload.get("gpu_model"),
            tpu=tpu,
            scheduler_enabled=payload.get("scheduler_enabled", False),
            preemptible_node=payload.get("preemptible_node", False),
            resource_affinity=payload.get("resource_affinity", ()),
        )

    def convert_tpu_preset(self, payload: dict[str, Any]) -> TPUPreset:
        return TPUPreset(
            type=payload["type"], software_version=payload["software_version"]
        )

    def convert_idle_job(self, payload: dict[str, Any]) -> IdleJobConfig:
        return IdleJobConfig(
            count=payload["count"],
            image=payload["image"],
            image_secret=payload.get("image_secret", ""),
            resources=self.convert_resources(payload["resources"]),
            env=payload.get("env") or {},
            node_selector=payload.get("node_selector") or {},
        )

    def convert_resources(self, payload: dict[str, Any]) -> Resources:
        return Resources(
            cpu_m=payload["cpu_m"],
            memory_mb=payload["memory_mb"],
            gpu=payload.get("gpu", 0),
        )

    def convert_storage(self, payload: dict[str, Any]) -> StorageConfig:
        return StorageConfig(
            url=URL(payload["url"]),
            volumes=[self.convert_volume(e) for e in payload.get("volumes", ())],
        )

    def convert_volume(self, payload: dict[str, Any]) -> VolumeConfig:
        return VolumeConfig(path=payload.get("path"), size_mb=payload.get("size_mb"))

    def convert_blob_storage(self, payload: dict[str, Any]) -> BlobStorageConfig:
        return BlobStorageConfig(url=URL(payload["url"]))

    def convert_registry(self, payload: dict[str, Any]) -> RegistryConfig:
        return RegistryConfig(url=URL(payload["url"]))

    def convert_monitoring(self, payload: dict[str, Any]) -> MonitoringConfig:
        return MonitoringConfig(url=URL(payload["url"]))

    def convert_secrets(self, payload: dict[str, Any]) -> SecretsConfig:
        return SecretsConfig(url=URL(payload["url"]))

    def convert_metrics(self, payload: dict[str, Any]) -> MetricsConfig:
        return MetricsConfig(url=URL(payload["url"]))

    def convert_dns(self, payload: dict[str, Any]) -> DNSConfig:
        return DNSConfig(
            name=payload["name"],
            a_records=[self.convert_a_record(r) for r in payload.get("a_records", [])],
        )

    def convert_a_record(self, payload: dict[str, Any]) -> ARecord:
        return ARecord(
            name=payload["name"],
            ips=payload.get("ips", []),
            dns_name=payload.get("dns_name", ARecord.dns_name),
            zone_id=payload.get("zone_id", ARecord.zone_id),
            evaluate_target_health=payload.get(
                "evaluate_target_health", ARecord.evaluate_target_health
            ),
        )

    def convert_disks(self, payload: dict[str, Any]) -> DisksConfig:
        return DisksConfig(
            url=URL(payload["url"]),
            storage_limit_per_user_gb=payload["storage_limit_per_user_gb"],
        )

    def convert_buckets(self, payload: dict[str, Any]) -> BucketsConfig:
        return BucketsConfig(
            url=URL(payload["url"]),
            disable_creation=payload.get("disable_creation", False),
        )

    def convert_ingress(self, payload: dict[str, Any]) -> IngressConfig:
        return IngressConfig(
            acme_environment=ACMEEnvironment(payload["acme_environment"]),
            cors_origins=payload.get("cors_origins", ()),
        )
